# SPDX-License-Identifier: MIT
# Copyright (c) 2026 LlamaIndex Inc.
"""Annotated types for Pydantic serialization of tricky tick/result fields.

Provides custom serializers/validators for:
- Events (polymorphic Pydantic models)
- Exceptions (not natively serializable)
- Event types (type[Event] as qualified name strings)
"""

from __future__ import annotations

from typing import Annotated, Any, Optional

from pydantic import PlainSerializer, PlainValidator
from workflows.context.serializers import JsonSerializer
from workflows.context.utils import (
    import_module_from_qualified_name,
)
from workflows.events import Event

_json_serializer = JsonSerializer()


def _serialize_event(event: Event) -> Any:
    return _json_serializer.serialize_value(event)


def _deserialize_event(data: Any) -> Event:
    return _json_serializer.deserialize_value(data)


SerializableEvent = Annotated[
    Event,
    PlainSerializer(_serialize_event, return_type=Any),
    PlainValidator(_deserialize_event),
]


def _serialize_optional_event(event: Event | None) -> Any:
    if event is None:
        return None
    return _json_serializer.serialize_value(event)


def _deserialize_optional_event(data: Any) -> Event | None:
    if data is None:
        return None
    return _json_serializer.deserialize_value(data)


SerializableOptionalEvent = Annotated[
    Optional[Event],
    PlainSerializer(_serialize_optional_event, return_type=Any),
    PlainValidator(_deserialize_optional_event),
]


def _serialize_exception(exc: Exception) -> dict[str, Any]:
    exc_type = type(exc)
    qualified_name = f"{exc_type.__module__}.{exc_type.__qualname__}"
    return {
        "exception_type": qualified_name,
        "exception_message": str(exc),
    }


def _deserialize_exception(data: Any) -> Exception:
    if isinstance(data, Exception):
        return data
    exc_message = data["exception_message"]
    try:
        exc_cls = import_module_from_qualified_name(data["exception_type"])
        return exc_cls(exc_message)
    except (ImportError, AttributeError, ValueError):
        return Exception(exc_message)


SerializableException = Annotated[
    Exception,
    PlainSerializer(_serialize_exception, return_type=dict[str, Any]),
    PlainValidator(_deserialize_exception),
]


def _serialize_event_type(event_type: type[Event]) -> str:
    return f"{event_type.__module__}.{event_type.__qualname__}"


def _deserialize_event_type(data: Any) -> type[Event]:
    if isinstance(data, type):
        return data
    return import_module_from_qualified_name(data)


SerializableEventType = Annotated[
    type[Event],
    PlainSerializer(_serialize_event_type, return_type=str),
    PlainValidator(_deserialize_event_type),
]



# SPDX-License-Identifier: MIT
# Copyright (c) 2026 LlamaIndex Inc.
"""
Ticks (events) that drive the control loop.

The control loop waits for ticks to arrive, then processes them through a reducer
to produce updated state and commands. Ticks represent all the different kinds of
events that can occur during workflow execution:
  - New events added to the workflow
  - Step function execution completing
  - Timeout occurring
  - User cancellation
  - External event publishing requests
"""

from __future__ import annotations

from typing import Annotated, Literal, Union

from pydantic import BaseModel, ConfigDict, Discriminator, TypeAdapter
from workflows.runtime.types.results import StepFunctionResult
from workflows.runtime.types.serialization_helpers import SerializableEvent


class TickStepResult(BaseModel):
    """When processed, executes a step function and publishes the result"""

    model_config = ConfigDict(frozen=True, arbitrary_types_allowed=True)
    type: Literal["step_result"] = "step_result"
    step_name: str
    worker_id: int
    event: SerializableEvent
    result: list[Annotated[StepFunctionResult, Discriminator("type")]]


class TickAddEvent(BaseModel):
    """When sent, adds an event to the workflow's event queue"""

    model_config = ConfigDict(frozen=True, arbitrary_types_allowed=True)
    type: Literal["add_event"] = "add_event"
    event: SerializableEvent
    step_name: str | None = None
    attempts: int | None = None
    first_attempt_at: float | None = None


class TickCancelRun(BaseModel):
    """When processed, cancels the workflow run"""

    model_config = ConfigDict(frozen=True)
    type: Literal["cancel_run"] = "cancel_run"


class TickPublishEvent(BaseModel):
    """When sent, publishes an event to workflow consumers, e.g. a UI or a callback"""

    model_config = ConfigDict(frozen=True, arbitrary_types_allowed=True)
    type: Literal["publish_event"] = "publish_event"
    event: SerializableEvent


class TickTimeout(BaseModel):
    """When processed, times the workflow out, cancelling it"""

    model_config = ConfigDict(frozen=True)
    type: Literal["timeout"] = "timeout"
    timeout: float


class TickIdleCheck(BaseModel):
    """Scheduled after state appears idle, to re-check after async events drain.

    Appended to tick_buffer when the reducer sees quiescent state. Processed
    on the next loop iteration after asyncio.sleep(0), giving in-flight
    ctx.send_event() calls a chance to deliver via the pull task.
    """

    model_config = ConfigDict(frozen=True)
    type: Literal["idle_check"] = "idle_check"


WorkflowTick = Annotated[
    Union[
        TickStepResult,
        TickAddEvent,
        TickCancelRun,
        TickPublishEvent,
        TickTimeout,
        TickIdleCheck,
    ],
    Discriminator("type"),
]

WorkflowTickAdapter: TypeAdapter[WorkflowTick] = TypeAdapter(WorkflowTick)
